package info.textgrid.lab.collatex.ui.parts;

import info.textgrid.lab.collatex.CollateXPlugin;
import info.textgrid.lab.collatex.model.equivalences.EToken;
import info.textgrid.lab.collatex.model.equivalences.Equivalence;
import info.textgrid.lab.collatex.model.equivalences.Equivalences;
import info.textgrid.lab.collatex.model.equivalences.Position;
import info.textgrid.lab.collatex.model.equivalences.TokenIndex;
import info.textgrid.lab.collatex.ui.CollateXPerspective;
import info.textgrid.lab.collatex.ui.WitnessTransfer;
import info.textgrid.lab.collatex.ui.utils.JobWithUICleanup;
import info.textgrid.lab.collatex.ui.utils.NotificationViewerElement;
import info.textgrid.lab.core.model.TextGridObject;
import info.textgrid.lab.core.swtutils.AdapterUtils;
import info.textgrid.lab.core.swtutils.CoreExceptionStatus;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Set;

import javax.xml.bind.DataBindingException;
import javax.xml.bind.JAXB;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IStorage;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.layout.TableColumnLayout;
import org.eclipse.jface.viewers.BaseLabelProvider;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.CheckboxCellEditor;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ColumnPixelData;
import org.eclipse.jface.viewers.DelegatingStyledCellLabelProvider;
import org.eclipse.jface.viewers.DelegatingStyledCellLabelProvider.IStyledLabelProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StyledString;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerDropAdapter;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.dnd.TransferData;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IReusableEditor;
import org.eclipse.ui.IStorageEditorInput;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.WorkbenchException;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.progress.UIJob;
import org.eclipse.ui.statushandlers.StatusManager;

import com.google.common.io.FileBackedOutputStream;

import eu.interedition.collatex.Token;

/**
 * Editor for a normalization
 * 
 * @see Equivalences
 */
public class RegularisationEditor extends EditorPart implements
IReusableEditor, IResourceChangeListener {

	private final class TokenColumnLabelProvider extends BaseLabelProvider
	implements IStyledLabelProvider {

		@Override
		public StyledString getStyledText(final Object element) {
			if (element instanceof NotificationViewerElement)
				return ((NotificationViewerElement) element).getStyledText();

			final StyledString result = new StyledString();
			if (element instanceof Equivalence) {
				final Equivalence equivalence = (Equivalence) element;
				final Iterator<EToken> iterator = ((Equivalence) element)
						.getTokens().iterator();
				while (iterator.hasNext()) {
					final EToken eToken = iterator.next();
					result.append(eToken.getNormalizedContent());
					if (equivalence.isLocal() && eToken.getPosition() != null) {
						final Position position = eToken.getPosition();
						if (position instanceof TokenIndex) {
							final TokenIndex pos = (TokenIndex) position;
							result.append(MessageFormat.format(" ({0}:{1})", //$NON-NLS-1$
									formatURI(pos.getUri()), pos.getIndex()),
									StyledString.COUNTER_STYLER);
						} else
							result.append(formatURI(position.getUri()),
									StyledString.COUNTER_STYLER);
					}
					if (iterator.hasNext())
						result.append(", "); //$NON-NLS-1$
				}
			}
			return result;
		}

		private String formatURI(final URI uri) {
			if (TextGridObject.SCHEMA.equals(uri.getScheme()))
				try {
					final TextGridObject textGridObject = TextGridObject
							.getInstance(uri, true);
					final String title = textGridObject.getTitle();
					if (title != null)
						return title;
				} catch (final CoreException e) {
					StatusManager.getManager().handle(e,
							CollateXPlugin.PLUGIN_ID);
				}
			return uri.toString();
		}

		@Override
		public Image getImage(final Object element) {
			return null;
		}
	}

	public static final String ID = "info.textgrid.lab.collatex.ui.parts.RegularisationEditor"; //$NON-NLS-1$
	private Table table;
	private Equivalences equivalences; // = new Equivalences();
	private boolean readOnly;
	private boolean dirty;
	private TableViewer equivalencesViewer;
	private IFile file;

	public RegularisationEditor() {
	}

	/**
	 * Create contents of the editor part.
	 * 
	 * @param parent
	 */
	@Override
	public void createPartControl(final Composite parent) {
		final Composite container = new Composite(parent, SWT.NONE);
		container.setLayout(new GridLayout(1, false));

		final Composite composite = new Composite(container, SWT.NONE);
		composite.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1,
				1));
		final TableColumnLayout tcl_composite = new TableColumnLayout();
		composite.setLayout(tcl_composite);

		equivalencesViewer = new TableViewer(composite, SWT.BORDER
				| SWT.FULL_SELECTION);
		table = equivalencesViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		final TableViewerColumn tokenColumn = new TableViewerColumn(
				equivalencesViewer, SWT.NONE);
		final TableColumn tokensTableColumn = tokenColumn.getColumn();
		tokensTableColumn
		.setToolTipText(Messages.RegularisationEditor_TokensColumnTooltip);
		tcl_composite.setColumnData(tokensTableColumn, new ColumnPixelData(300,
				true, true));
		tokensTableColumn
		.setText(Messages.RegularisationEditor_TokenColumnsTitle);
		tokenColumn.setLabelProvider(new DelegatingStyledCellLabelProvider(
				new TokenColumnLabelProvider()));

		equivalencesViewer.addDropSupport(DND.DROP_COPY | DND.DROP_LINK
				| DND.DROP_MOVE,
				new Transfer[] { WitnessTransfer.getInstance() },
				new ViewerDropAdapter(equivalencesViewer) {

			@Override
			public boolean validateDrop(final Object target,
					final int operation, final TransferData transferType) {
				//				System.err.println("Validate validate validate ..." //$NON-NLS-1$
				// + transferType.type);
				return WitnessTransfer.getInstance().isSupportedType(
						transferType);
			}

			@Override
			public boolean performDrop(final Object data) {
				final Equivalence target = getCurrentTarget() instanceof Equivalence ? (Equivalence) getCurrentTarget()
						: null;
				Equivalence equivalence;
				if (getCurrentLocation() == LOCATION_ON
						&& target != null)
					equivalence = target;
				else {
					equivalence = new Equivalence();
					equivalences.getEquivalences().add(equivalence);
				}
				for (final Token token : WitnessTransfer.getInstance()
						.getTokens())
					equivalence.getTokens().add(new EToken(token));
				getViewer().refresh();
				setDirty(true);

				// XXX Debugging:
				StatusManager
				.getManager()
				.handle(new Status(
						IStatus.INFO,
						CollateXPlugin.PLUGIN_ID,
						MessageFormat
						.format("Regularisation Editor: Drop performed.\n  Equivalence: {0}\n  Tokens: {1}\n  Target: {2}", //$NON-NLS-1$
								equivalence,
								WitnessTransfer
								.getInstance()
								.getTokens(),
								target)));

				return true;
			}
		});

		final TableViewerColumn localColumn = new TableViewerColumn(
				equivalencesViewer, SWT.NONE);
		final TableColumn localTableColumn = localColumn.getColumn();
		localTableColumn
		.setToolTipText(Messages.RegularisationEditor_LocalColumnTooltip);
		tcl_composite.setColumnData(localTableColumn, new ColumnPixelData(150,
				true, true));
		localTableColumn
		.setText(Messages.RegularisationEditor_LocalColumnTitle);

		localColumn.setLabelProvider(new ColumnLabelProvider() {

			@Override
			public String getText(final Object element) {
				if (element instanceof Equivalence)
					return ((Equivalence) element).isLocal() ? Messages.RegularisationEditor_local
							: Messages.RegularisationEditor_global;
				else
					return null;
			}

		});

		localColumn.setEditingSupport(new EditingSupport(equivalencesViewer) {

			private CheckboxCellEditor checkboxCellEditor;

			@Override
			protected void setValue(final Object element, final Object value) {
				if (element instanceof Equivalence && value instanceof Boolean) {
					((Equivalence) element).setLocal((Boolean) value);
					getViewer().update(element, null);
					setDirty(true);
				} else
					throw new IllegalArgumentException();
			}

			@Override
			protected Object getValue(final Object element) {
				if (element instanceof Equivalence)
					return Boolean.valueOf(((Equivalence) element).isLocal());
				else
					throw new IllegalArgumentException();
			}

			@Override
			protected CellEditor getCellEditor(final Object element) {
				if (checkboxCellEditor == null)
					checkboxCellEditor = new CheckboxCellEditor(null, SWT.CHECK);
				return checkboxCellEditor;
			}

			@Override
			protected boolean canEdit(final Object element) {
				return element instanceof Equivalence;
			}
		});

		final Label lblDropTokensFrom = new Label(container, SWT.WRAP);
		lblDropTokensFrom.setLayoutData(new GridData(SWT.FILL, SWT.CENTER,
				false, false, 1, 1));
		lblDropTokensFrom.setText(Messages.RegularisationEditor_MiniHowto);

		equivalencesViewer.setContentProvider(new IStructuredContentProvider() {

			@Override
			public void inputChanged(final Viewer viewer,
					final Object oldInput, final Object newInput) {
			}

			@Override
			public void dispose() {
			}

			@Override
			public Object[] getElements(final Object inputElement) {
				if (inputElement instanceof Equivalences) {
					final Set<Equivalence> equivalences = ((Equivalences) inputElement)
							.getEquivalences();
					if (equivalences.isEmpty())
						return new Object[] { new NotificationViewerElement(
								Messages.RegularisationEditor_HowtoNotice) };
					return equivalences.toArray();
				} else if (inputElement instanceof NotificationViewerElement)
					return new Object[] { inputElement };
				return null;
			}
		});

		if (equivalences != null)
			equivalencesViewer.setInput(equivalences);
		else
			equivalencesViewer.setInput(new NotificationViewerElement(
					Messages.RegularisationEditor_LoadingNotice));

		PlatformUI
		.getWorkbench()
		.getHelpSystem()
		.setHelp(parent,
				"info.textgrid.lab.collatex.regularisationEditor"); //$NON-NLS-1$
	}

	@Override
	public void setFocus() {
		// Set the focus
	}

	@Override
	public void doSave(final IProgressMonitor monitor) {
		final SubMonitor progress = SubMonitor.convert(monitor, 100);
		if (getEditorInput() instanceof IFileEditorInput) {
			final IFile file = ((IFileEditorInput) getEditorInput()).getFile();
			progress.worked(5);
			if (file.isReadOnly()) {
				progress.setCanceled(true);
				// FIXME display error?
				return;
			}
			final FileBackedOutputStream outputStream = new FileBackedOutputStream(
					1024 * 1024 * 1024);
			JAXB.marshal(equivalences, outputStream);
			progress.worked(45);
			try {
				file.setContents(outputStream.getSupplier().getInput(), true,
						true, progress.newChild(50));
				new UIJob("") { //$NON-NLS-1$

					@Override
					public IStatus runInUIThread(final IProgressMonitor monitor) {
						setDirty(false);
						return Status.OK_STATUS;
					}

				}.schedule();
			} catch (final CoreException e) {
				StatusManager.getManager().handle(e, CollateXPlugin.PLUGIN_ID);
				progress.setCanceled(true);
			} catch (final IOException e) {
				StatusManager
				.getManager()
				.handle(new Status(
						IStatus.ERROR,
						CollateXPlugin.PLUGIN_ID,
						"An IO error occured while reading from an internal stream. Should. Not. Happen.", //$NON-NLS-1$
						e));
				progress.setCanceled(true);
			}
			progress.done();
		} else
			progress.setCanceled(true);
	}

	@Override
	public void doSaveAs() {
		// Do the Save As operation
	}

	@Override
	public void init(final IEditorSite site, final IEditorInput input)
			throws PartInitException {
		setSite(site);
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this,
				IResourceChangeEvent.POST_CHANGE);
		if (input instanceof IStorageEditorInput)
			setInput(input);
		else
			throw new PartInitException(Messages.RegularisationEditor_12);

		try {
			PlatformUI.getWorkbench().showPerspective(CollateXPerspective.ID,
					site.getWorkbenchWindow());
		} catch (final WorkbenchException e) {
			StatusManager.getManager().handle(e, CollateXPlugin.PLUGIN_ID);
		}
	}

	@Override
	public void setInput(final IEditorInput input) {
		if (input instanceof IStorageEditorInput) {
			super.setInput(input);
			if (input instanceof IFileEditorInput)
				this.file = ((IFileEditorInput) input).getFile();
			else
				this.file = null;

			new JobWithUICleanup(Messages.RegularisationEditor_Loading) {

				private IStorage storage;
				private boolean dirty = true;

				@Override
				protected IStatus run(final IProgressMonitor monitor) {
					final SubMonitor progress = SubMonitor.convert(monitor, 10);
					try {
						storage = ((IStorageEditorInput) input).getStorage();
						progress.worked(3);
						if (input.exists()) {
							final TextGridObject textGridObject = AdapterUtils
									.getAdapter(storage, TextGridObject.class);
							if (textGridObject != null
									&& textGridObject.isNew()) {
								this.dirty = true;
								equivalences = new Equivalences();
								progress.worked(7); // nothing to read.
							} else {
								final InputStream inputStream = storage
										.getContents();
								progress.worked(3);
								try {
									equivalences = JAXB.unmarshal(inputStream,
											Equivalences.class);
									dirty = false;
								} catch (final DataBindingException e) {
									return new Status(
											IStatus.WARNING,
											CollateXPlugin.PLUGIN_ID,
											NLS.bind(
													Messages.RegularisationEditor_Failed_to_parse,
													storage), e);
								}
							}
							progress.worked(4);
						}

						progress.done();
						return Status.OK_STATUS;
					} catch (final CoreException e) {
						progress.done();
						return new CoreExceptionStatus(
								CollateXPlugin.PLUGIN_ID, e,
								Messages.RegularisationEditor_Failed_to_load);
					}
				}

				@Override
				protected IStatus done(final IJobChangeEvent event,
						final IProgressMonitor monitor) {

					if (!event.getResult().isOK())
						if (event.getResult().getException() != null
						&& event.getResult().getException() instanceof DataBindingException) {
							equivalences = new Equivalences();
							ErrorDialog
							.openError(
									getSite().getShell(),
									Messages.RegularisationEditor_ParseErrorDlg_Title,
									NLS.bind(
											Messages.RegularisationEditor_ParseErrorDlg_Message,
											input.getName()), event
											.getResult(), 0xFFFF);
						}

					setPartName(input.getName());
					readOnly = storage.isReadOnly();
					setDirty(dirty);
					setInputWithNotify(input);

					if (equivalences != null && equivalencesViewer != null
							&& !equivalencesViewer.getControl().isDisposed())
						equivalencesViewer.setInput(equivalences);

					return Status.OK_STATUS;
				}
			}.schedule();

		}

		// } catch (final DataBindingException e) {
		// this.setDirty(true);
		// final StatusAdapter statusAdapter = new StatusAdapter(
		// new Status(
		// IStatus.ERROR,
		// CollateXPlugin.PLUGIN_ID,
		// MessageFormat
		// .format("Failed to parse equivalence set stream: {0}",
		// e.getLocalizedMessage()), e));
		// statusAdapter.setProperty(
		// IStatusAdapterConstants.TITLE_PROPERTY, getPartName());
		// statusAdapter
		// .setProperty(
		// IStatusAdapterConstants.EXPLANATION_PROPERTY,
		// "The contents of this editor's input has not been understood, and thus the editor has been initialized with empty contents. If you modify and save it, the original input will be overwritten.");
		// statusAdapter.setProperty(
		// IStatusAdapterConstants.HINT_PROPERTY,
		// "Close the editor without saving and try to open the input with the XML editor if you think something went wrong.");
		// StatusManager.getManager().handle(statusAdapter,
		// StatusManager.SHOW);
		// } catch (final CoreException e) {
		// throw new PartInitException("Failed to read input.", e);
		// }
		// }
	}

	@Override
	public boolean isDirty() {
		return dirty;
	}

	@Override
	public boolean isSaveAsAllowed() {
		return false;
	}

	public void setDirty(final boolean dirty) {
		this.dirty = dirty;
		firePropertyChange(PROP_DIRTY);
	}

	@Override
	public void resourceChanged(final IResourceChangeEvent event) {
		if (file != null) {
			final IResourceDelta delta = event.getDelta().findMember(
					file.getFullPath());
			if (delta != null)
				if ((delta.getFlags() & IResourceDelta.MOVED_TO) != 0) {
					final IFile newFile = (IFile) ResourcesPlugin.getWorkspace().getRoot().findMember(delta.getMovedToPath());
					new UIJob(Messages.RegularisationEditor_AdaptingMove) {

						@Override
						public IStatus runInUIThread(
								final IProgressMonitor monitor) {
							setInput(new FileEditorInput(newFile));
							return Status.OK_STATUS;
						}

					}.schedule();
				}
		}
	}

	public void deleteSelection() {
		if (equivalencesViewer != null
				&& !equivalencesViewer.getControl().isDisposed()) {
			final IStructuredSelection selection = (IStructuredSelection) equivalencesViewer
					.getSelection();
			if (!selection.isEmpty()) {
				final Equivalence[] selectedEquivs = AdapterUtils.getAdapters(
						selection.toArray(), Equivalence.class, false);
				equivalences.getEquivalences().removeAll(
						Arrays.asList(selectedEquivs));
				equivalencesViewer.remove(selectedEquivs);
			}
		}
	}
}